#ifndef TRANSCIEVER_H
#define TRANSCIEVER_H

#include <map>
#include <ctime>
#include <chrono>
#include <thread>
#include <numeric>
#include <algorithm>

#include "device.h"

/* Parameters vector to display */
const int VENDOR_PARAMS = 0;
const int MODULE_PARAMS = 1;

/* How to save the parameters to transciever */
const int SAVE_PERMANENT = 0;
const int SAVE_TEMPORARILY = 1;

/* Modes supported by the transciever */
const int MODE_NORMAL = 0;		// Transmittion of data both directions
const int MODE_WAKE_UP = 1;		// Preamble added to the data to wake up the other device
const int MODE_POWER_DOWN = 2;	// Unable to transmit. Need to receive the wake up preamble.
const int MODE_SLEEP = 3; 		// Used for setting up the parameters of the transmittion

/* Parity bits */
#define PARITY_BIT_8N1 0b00		// (default)
#define PARITY_BIT_8O1 0b01
#define PARITY_BIT_8E1 0b11

/* Baud rate settings */
#define BAUD_1200 0b000			// 1200 baud
#define BAUD_2400 0b001			// 2400 baud
#define BAUD_4800 0b010			// 4800 baud
#define BAUD_9600 0b011			// 9600 baud (default)
#define BAUD_19200 0b100		// 19200 baud
#define BAUD_38400 0b101		// 34800 baud
#define BAUD_57600 0b110		// 57600 baud
#define BAUD_115200 0b111		// 115200 baud

/* Air data rate */
#define AIR_03K 0b000			// 0.3k 
#define AIR_12K 0b001			// 1.2k
#define AIR_24K 0b010			// 2.4k (default)
#define AIR_48K 0b011			// 4.8k 
#define AIR_96K 0b100			// 9.6k
#define AIR_192K 0b101			// 19.2k 
// options 0b110 and 0b111 are same as 19.2k

/* Parameters to get or set */
const int PARAM_ADDRESS = 0;
const int PARAM_PARITY = 1;
const int PARAM_UART_DATARATE = 2;
const int PARAM_AIR_DATARATE = 3;

class Transciever : public Device 
{
public:
	// Used for hardware check of the connected device
	int init_device(std::string full_path);			
	// Generates the lookup table for the baud rate (intended for later stages)					
	void generate_lookup();
	// Search for message to send and clearing old messages from buffer
	void start_from_message(const std::string message = "");
	
	// Getters
	void get_vendor_data();
	void get_module_parameters();
	bool data_available();
	uint16_t get_parameter(const int parameter);
	
	uint8_t get_counter_in()
			{return __counter_in;}
	
	std::string get_error_message()
			{return __error_message;}

	int number_of_messages_available()
			{return __received_messages.size();}

	std::string get_message();
	
	// Setters
	void set_mode(const int mode);
	void set_parameter(const int parameter, int16_t value, const int mode);
	

	// Communication
	int receive_data();
	// Generating 16 bit CRC code
	uint16_t gencrc_16(const std::string & message);
	// Processing received message (CRC, rolling counter check)
	int process_received_data();
	// Construction of the packet
	void process_data_to_send(std::string & command, TRANSMITTION_PRIORITY priority);
	bool send_data();

	// Closing connection
	void terminate_connection();
	

private:
	uint8_t __counter_in;
	uint8_t __counter_out;
	uint8_t __pin_M0;
	uint8_t __pin_M1;
	uint8_t __pin_AUX;
	std::string __full_path;
	std::vector<uint8_t> __vendor_parameters;
	std::vector<uint8_t> __transmittion_parameters;
	std::deque<std::tuple<uint8_t,std::string>> __messages_to_send;
	std::deque<std::string> __received_messages;
	std::map<const int,enum BAUD_RATE> __lookup_table;
	/* 
	 * Setting of parameters of the tranceiver always has to be done at 9600
	 * so when other intended for use (faster) has to be stored somewhere else
	 * while parameters are set and then the communication reopens with higher BAUD
	 */
	enum BAUD_RATE __rate;
	enum TRANSMITTION_STATE __state;
private:
	// Check of the AUX logic level
	bool device_is_ready();
	// Adding of the message to the output buffer. If it is DATA< then to the end, if it is COMMAND, then to the front
	void set_message(std::tuple<uint8_t,std::string> message, TRANSMITTION_PRIORITY priority)
		{ (priority == DATA) ? 	__messages_to_send.push_back(message):
							 	__messages_to_send.push_front(message);}
};
#endif
