#include "../include/gps.h"


/*
 * Initialization of the GPS devices connected
 * First the regex expression is assigned
 * to private variable regex_key. Then we
 * try to open the port. If succesful we read the first packet 
 * and then the initialization is complete 
 */
int GPS::init_device(std::string path)
{
	__regex_key = GPS_PACKET;

	int port_status = open_port(path, COM_BAUD_9600);
	if(port_status != ERROR_SUCCESS)
		return ERROR_SERIAL_PORT_CLOSED;

	while(!data_available());
	if(Device::receive_data() > 0)
		return ERROR_SUCCESS;

	return ERROR_DEVICE_NOT_RECOGNIZED;
}

/*
 * First we make the regex search in the __buffer of received bytes
 * If succesful, we log the raw packet into the log file and 
 * extract data for transfer.
 */
int GPS::process_received_data()
{
	std::smatch match;
	std::regex reg(__regex_key);
	std::regex_search(__buffer,match,reg);
	std::string data = match.str();

#ifdef DEBUG_GPS
	std::cout << "GPS received: " << __buffer << std::endl;
	std::cout << "Size of buffer: "  << __buffer.size() << std::endl;
	std::cout << "Extracted message: " << data << std::endl;
#endif

	__buffer.clear();

	if(match.str().size()  >= GPS_MIN_LENGTH &&
		extract_data_to_transfer(data))
	{
		__to_log = match.str();
		__to_transfer = data;
		return ERROR_SUCCESS;
	}
	return ERROR_DATA_NOT_RECOGNIZED;
}

/*
 * GNGGA specific extraction of the data from packet
 * When other format will be used later, this function
 * has to be updated to fit with what extraction is intended
 */

bool GPS::extract_data_to_transfer(std::string & message)
{
	std::stringstream separator(message);
	std::string segment;
	std::vector<std::string> segment_list;
	std::vector<uint8_t> output;

	while(std::getline(separator,segment,','))
		segment_list.push_back(segment);

	segment_list.erase(segment_list.begin());
	message.clear();

	if(segment_list.size() >= 5)
	{
		float sign = 1;
		convert_number_to_byte(segment_list[0],output,sign); // Timestamp encoding
		sign = (segment_list[2] == "N") ? 1 : -1;
		convert_number_to_byte(segment_list[1],output,sign); // Latitude encoding
		sign = (segment_list[4] == "E") ? 1 : -1;
		convert_number_to_byte(segment_list[3],output,sign); // Longitude encoding
	}
	message.assign(output.begin(),output.end());

#ifdef DEBUG
	std::cout << "GPS message to transfer coded: ";
	for(size_t i = 0; i < message.size(); i++)
		printf("%x",message[i]);
	std::cout << std::endl;	
#endif

	return (message.length() >= GPS_MIN_LENGTH) ? true : false;
}


/*
 * Conversion of the string representation of the numbers to the byte one.
 */
void GPS::convert_number_to_byte(std::string message, std::vector<uint8_t> & output, float sign)
{
	float number = sign * std::stof(message);
	uint8_t const *pointer = reinterpret_cast<uint8_t const *>(&number);
	for(size_t i =0; i < sizeof(float); i++)
		output.push_back(pointer[i]);	
}

void GPS::terminate_connection()
{
	serialClose(__file_descriptor);
}
