#include "../include/receiver.h"

// Shared variables by all threads
std::string command;
std::deque<std::string> data_queue;  
std::mutex cmd_mutex;
std::mutex dat_mutex;

union {
	float number;
	uint8_t bytes[4];
} convertor;

// Thread for receiving commands from Python
void command_thread()
{
#ifdef DEBUG
	std::cout << "commnand thread started" << std::endl;
#endif
	const char *fifo_cmd = "fifo_cmd";

	std::string cmd;
	mknod(fifo_cmd, S_IFIFO | 0666, 0);

	while(cmd != "stop")
	{
		std::ifstream cmd_stream(fifo_cmd);
		if(!cmd_stream.good())
			std::cerr << "problem opening command fifo" << std::endl;

		getline(cmd_stream,cmd);
		set_command(cmd);
		cmd_stream.close();
	}
	
	std::cout << "closing command thread" << std::endl;
#ifdef DEBUG
	std::cout << "command  thread terminated" << std::endl;
#endif		
}

// Thread for transmitting data to Python
void data_thread()
{
#ifdef DEBUG	
	std::cout << "starting data thread" << std::endl;
#endif
	const char *fifo_dat = "fifo_dat";

	std::string data;
	mknod(fifo_dat, S_IFIFO | 0666,0);
	while(data != "None")
	{
		data.clear();
		while(data == "")
		{
			data = get_data();
			std::this_thread::sleep_for (std::chrono::milliseconds(5));
		}
		std::cout << "got data: " << data << std::endl;
		std::ofstream dat(fifo_dat);

		if(!dat.good())
			std::cerr << "error opening data fifo" << std::endl;

		dat << data;
		dat.close();
	}
#ifdef DEBUG	
	std::cout << "closing data thread" << std::endl;
#endif
}

void set_command(std::string t_command)
{
	std::lock_guard<std::mutex> lock(cmd_mutex);
	command = t_command;
}

void set_data(std::string t_data)
{
	std::lock_guard<std::mutex> lock(dat_mutex);
	data_queue.push_back(t_data);
}

std::string get_command()
{
	std::lock_guard<std::mutex> lock(cmd_mutex);
	std::string cmd = command;
	command.clear();
	return cmd;
}

std::string get_data()
{
	std::lock_guard<std::mutex> lock(dat_mutex);
	std::string tmp;
	if(!data_queue.empty())
	{
		tmp = data_queue.front();
		data_queue.pop_front();
	}
	return tmp;
}

bool data_empty()
{
	std::lock_guard<std::mutex> lock(dat_mutex);
	return data_queue.empty();
}

/************************** RECEIVER **************************/

/*
 * Initialization of the library for UART and GPIO called wiringPi
 * Further on we initialize the transceiver module
 */
int Receiver::init_devices()
{
	// Setups the library
	if(wiringPiSetup() == -1)
	{
		__error_message = std::strerror (errno);
		return ERROR_INIT_FAILED;
	}

#ifdef DEBUG
	std::cout << "wiringPI initialized" << std::endl;
#endif

	std::string filename = DEV_FOLDER + TTY_SERIAL0;
	__transciever = new Transciever();
	if(__transciever->init_device(filename) != ERROR_SUCCESS)
		return ERROR_INIT_FAILED;

#ifdef DEBUG
	std::cout << "initialized transciever module" << std::endl;
#endif
	return ERROR_SUCCESS;
}

int Receiver::run_loop()
{
	std::cout << "TRANSCIEVER SUCCESSFULLY INITIALIZED" << std::endl;
	__state = RECEIVER_READY;
	bool operating = true;
	std::clock_t start = clock();
	std::string cmd;
	std::string data;

	while(operating)
	{
		cmd = get_command();
		if(!cmd.empty())
			__state = static_cast<RECEIVER_STATE>(react_on_command(cmd));

		switch(__state)
		{
			case RECEIVER_TRANSMIT:
				start = clock();
				cmd = SEND;
				process_data_to_send(cmd);
				if(__transciever->send_data())
					__state = RECEIVER_WAIT;
				break;
				
			case RECEIVER_RECEIVE:
				__transciever->receive_data();
				for(int i = 0; i <= __transciever->number_of_messages_available(); i++)
				{
					data = __transciever->get_message();
					if(process_received_data(data) == ERROR_SUCCESS)
						set_data(data);
				}
				__state = RECEIVER_TRANSMIT;
				break;
				
			case RECEIVER_RESET:
				init_devices();
				__state = RECEIVER_READY;
				break;
				
			case RECEIVER_WAIT:	
				if(((double)(clock() - start)) / CLOCKS_PER_SEC >= RECEIVER_DATA_TIMEOUT)
					__state = RECEIVER_TRANSMIT;
				if(__transciever->data_available())
					__state = RECEIVER_RECEIVE;
				break;
					
			case RECEIVER_TERMINATE:
				std::cout << "terminating the program" << std::endl;
				operating = false;
				break;
				
			default:
				break;
		}
	}
	
	set_data("None");
	std::this_thread::sleep_for(std::chrono::seconds(1));
	__transciever->terminate_connection();
	return ERROR_SUCCESS;
}

// Communication functions 
int Receiver::react_on_command(std::string cmd)
{
	handle_command(cmd);
	if(cmd.find(STOP) != std::string::npos)
		return RECEIVER_TERMINATE;
	if(cmd.find(LOG) != std::string::npos)
		return RECEIVER_TRANSMIT;
	if(cmd.find(RESTART) != std::string::npos)
		return RECEIVER_RESET;
	return __state;	
}


void Receiver::process_data_to_send(std::string & t_command)
{
	if(t_command.find(SEND) != std::string::npos)
		t_command = t_command + " " + std::to_string(__transciever->get_counter_in()); 
	__transciever->process_data_to_send(t_command, DATA);
}

void Receiver::handle_command(std::string cmd)
{
	bool got_response = false;
	std::clock_t start;
	
	while(!got_response)
	{
		start = clock();
		process_data_to_send(cmd);
		__transciever->send_data();

		while((((double)(clock() - start)) / CLOCKS_PER_SEC) < RECEIVER_COMMAND_TIMEOUT) // Custom timeout for commands. Can be changed in constants.h
		{
			if(__transciever->data_available())
			{
				__transciever->receive_data();
				if(__transciever->get_message().find(ACK) != std::string::npos)
				{
					set_data(ACK);
					return;
				}
			}
		}
	}
}

/*
 *  Reading the bytes of received body from the transceiver
  * and converting 4 consecutive into the float number
 */
int Receiver::process_received_data(std::string & data)
{
	std::string output;
	for(size_t i = 0; i < data.size(); i++)
	{
		convertor.bytes[i % 4] = data[i];
		if(i > 0 && i % 4 == 3)
			output += std::to_string(convertor.number) + ',';
	}
	data = output.substr(0,output.size()-1);
	return ERROR_SUCCESS;
}
