
#include "../include/transciever.h"

int Transciever::init_device(std::string full_path)
{
	__pin_M0 = 0;
	__pin_M1 = 2;
	__pin_AUX = 3;

	// Setup the pins
	pinMode(__pin_M0,OUTPUT);
	pinMode(__pin_M1,OUTPUT);
	pinMode(__pin_AUX,INPUT);

	// Setting up internal pull-up resistors on RPi
	pullUpDnControl(__pin_M0, PUD_UP);
	pullUpDnControl(__pin_M1, PUD_UP);

	__rate = COM_BAUD_115200;
	__full_path = full_path;
	__file_descriptor = -1;

	// To obtain default values
	get_vendor_data();
	get_module_parameters();

	// Changing of the parameters of the transmittion.
	generate_lookup();
	set_parameter(PARAM_AIR_DATARATE, AIR_48K, SAVE_TEMPORARILY);
	set_parameter(PARAM_UART_DATARATE,BAUD_115200,SAVE_TEMPORARILY);
	set_parameter(PARAM_ADDRESS,0x0, SAVE_TEMPORARILY);
	set_parameter(PARAM_PARITY,PARITY_BIT_8N1,SAVE_TEMPORARILY);
	get_module_parameters();

	__counter_in = 0;
	__counter_out = 0;


	clock_t start = clock();
	while(double(clock()-start)/CLOCKS_PER_SEC < 1)
		if(device_is_ready())
			return ERROR_SUCCESS;
	return ERROR_CONNECTION_TIMEOUT;
}

void Transciever::generate_lookup()
{
	__lookup_table.insert(std::pair<const int, BAUD_RATE>(BAUD_1200,COM_BAUD_1200));
	__lookup_table.insert(std::pair<const int, BAUD_RATE>(BAUD_2400, COM_BAUD_2400));
	__lookup_table.insert(std::pair<const int, BAUD_RATE>(BAUD_4800, COM_BAUD_4800));
	__lookup_table.insert(std::pair<const int, BAUD_RATE>(BAUD_9600, COM_BAUD_9600));
	__lookup_table.insert(std::pair<const int, BAUD_RATE>(BAUD_19200, COM_BAUD_19200));
	__lookup_table.insert(std::pair<const int, BAUD_RATE>(BAUD_38400, COM_BAUD_38400));
	__lookup_table.insert(std::pair<const int, BAUD_RATE>(BAUD_57600,COM_BAUD_57600));
	__lookup_table.insert(std::pair<const int, BAUD_RATE>(BAUD_115200,COM_BAUD_115200));
}

bool Transciever::device_is_ready()
{
	return (digitalRead(__pin_AUX) == HIGH) ? true : false;
}

bool Transciever::data_available()
{
	// calling parent class function
	return Device::data_available();
}

// Changing mode of the transceiver module
void Transciever::set_mode(const int mode)
{
	// Waiting for the device to finish all transmissions.
	while(!device_is_ready())
		std::this_thread::sleep_for (std::chrono::milliseconds(1));
	std::this_thread::sleep_for (std::chrono::milliseconds(2));
	
	switch(mode){
		case MODE_NORMAL:
			digitalWrite(__pin_M0, LOW);
			digitalWrite(__pin_M1, LOW);

			// Opening port at specified rate
			serialClose(__file_descriptor);
			open_port(__full_path, __rate);
			break;
		case MODE_WAKE_UP:
			digitalWrite(__pin_M0, HIGH);
			digitalWrite(__pin_M1, LOW);
			break;
		case MODE_POWER_DOWN:
			digitalWrite(__pin_M0, LOW);
			digitalWrite(__pin_M1, HIGH);
			break;
		case MODE_SLEEP:
			digitalWrite(__pin_M0, HIGH);
			digitalWrite(__pin_M1, HIGH);

			// Opening port at programming rate
			serialClose(__file_descriptor);
			open_port(__full_path, COM_BAUD_9600);
			break;
	}
	std::this_thread::sleep_for (std::chrono::milliseconds(20));
}

void Transciever::get_vendor_data()
{
	__vendor_parameters.clear();
	// Enter programming mode
	set_mode(MODE_SLEEP);
	serialPutchar(__file_descriptor,0xC3);
	serialPutchar(__file_descriptor,0xC3);
	serialPutchar(__file_descriptor,0xC3);
	std::this_thread::sleep_for (std::chrono::milliseconds(20));

#ifdef DEBUG_TRC
	std::cout << "Vendor data - available data size: " 
			  << serialDataAvail(__file_descriptor)
			  << std::endl;
#endif

	while(serialDataAvail(__file_descriptor))
	{
		__vendor_parameters.push_back(serialGetchar(__file_descriptor));
#ifdef DEBUG_TRC
	std::printf("vendor data: %x\n",__vendor_parameters.back());
#endif
	}

	set_mode(MODE_NORMAL);
}

void Transciever::get_module_parameters()
{
	__transmittion_parameters.clear();

	// Enter programming mode
	set_mode(MODE_SLEEP);
	serialPutchar(__file_descriptor,0xC1);
	serialPutchar(__file_descriptor,0xC1);
	serialPutchar(__file_descriptor,0xC1);
	std::this_thread::sleep_for (std::chrono::milliseconds(20));

#ifdef DEBUG_TRC
	std::cout << "transmittion data - available data size: "
	          << serialDataAvail(__file_descriptor)
	          << std::endl;
#endif

	while(serialDataAvail(__file_descriptor))
	{
		__transmittion_parameters.push_back(serialGetchar(__file_descriptor));
#ifdef DEBUG_TRC
	std::printf("transmittion params: %x\n", __transmittion_parameters.back());
#endif
	}

	set_mode(MODE_NORMAL);
}

// Terminating the connection. All parameters are reset to default
void Transciever::terminate_connection()
{
	while(!device_is_ready());
	set_mode(MODE_SLEEP);
	serialPutchar(__file_descriptor,0xC4);
	serialPutchar(__file_descriptor,0xC4);
	serialPutchar(__file_descriptor,0xC4);
	set_mode(MODE_NORMAL);
	serialClose(__file_descriptor);
}

// Extracting values of each of transmission parameters for the tranceiver
// See the manufacturer datasheet for details.
// This is just subset of all of them. Others we decided not to change
uint16_t Transciever::get_parameter(const int parameter)
{
	if(__transmittion_parameters.empty())
		return 0xffff;

	switch(parameter)
	{
		case PARAM_ADDRESS:
			return __transmittion_parameters.at(1) << 8 | __transmittion_parameters.at(2);
		case PARAM_PARITY:
			return (__transmittion_parameters.at(3) & 0xC0) >> 6;
		case PARAM_UART_DATARATE:
			return (__transmittion_parameters.at(3) & 0x38) >> 3;
		case PARAM_AIR_DATARATE:
			return (__transmittion_parameters.at(3) & 0x7);
		default:
			return std::numeric_limits<uint16_t>::max();
	}
}

// Setting of each of the useful parameters
// Based on mode we set these parameters either temporarily or permanently
void Transciever::set_parameter(const int parameter, int16_t value, const int mode)
{
	if(__transmittion_parameters.empty())
		__transmittion_parameters.resize(5);

	__transmittion_parameters.at(0) = (mode == SAVE_PERMANENT) ? 0xC0 : 0xC2;
	switch(parameter)
	{
		case PARAM_ADDRESS:
			__transmittion_parameters.at(1) = value >> 8;
			__transmittion_parameters.at(2) = value & 255;
			break;
		case PARAM_PARITY:
			__transmittion_parameters.at(3) =  (__transmittion_parameters.at(3) & 0x3f) | value << 6;
			break;
		case PARAM_UART_DATARATE:
			__rate = __lookup_table[value];
			__transmittion_parameters.at(3) = (__transmittion_parameters.at(3) & 0xC7) | value << 3;
			break;
		case PARAM_AIR_DATARATE:
			__transmittion_parameters.at(3) =  (__transmittion_parameters.at(3) & 0xf8) | value;
			break;
	}

	set_mode(MODE_SLEEP);

	for(auto i = __transmittion_parameters.begin(); i != __transmittion_parameters.end(); i++)
		serialPutchar(__file_descriptor,*i);

	set_mode(MODE_NORMAL);
}

bool Transciever::send_data()
{
	std::string message;
	// Check whether device is busy
	while(!device_is_ready())
		if(serialDataAvail(__file_descriptor) > 0)
		{
			// If transceiver receives message
			receive_data(); 
#ifdef RECEIVER	
			// Check whether we would not ask for same data which we already succesfully received
			start_from_message(); 
#endif
		}	
	if(!__messages_to_send.empty())
		message = std::get<1>(__messages_to_send[0]);
	else
		return false;
		
#ifdef DEBUG_TRC
	std::cout << "sending packet: " << message << std::endl;
#endif
	// Sending the data over UART
	write(__file_descriptor,message.c_str(),message.size());
	while(!device_is_ready());

#ifdef RECEIVER
	__messages_to_send.clear();
#endif
	return true;
}

// Receive data gets bytes from the GPIO UART buffer.
int Transciever::receive_data()
{
	int succesfully_transmitted = 0;
	uint8_t received_char;

	__state = (__buffer.find(PACKET_START) == 0) ? FIND_END : FIND_START;

	while(serialDataAvail(__file_descriptor) > 0)
	{
		read(__file_descriptor,&received_char,1);
		switch(__state)
		{
			case FIND_START:
				if(received_char == '#')
				{
					__buffer.push_back(received_char);
					if(__buffer.find(PACKET_START) == 0)
						__state = FIND_END;
				}
				break;

			case FIND_END:
				__buffer.push_back(received_char);
				if(__buffer.find(PACKET_END) == __buffer.size() - 2)
				{
					__state = FIND_START;
					succesfully_transmitted += process_received_data();
				}
				
				if(__buffer.rfind(PACKET_START) == __buffer.size() -2)
				{
					__buffer.resize(__buffer.size() - 2);
					succesfully_transmitted += process_received_data();
					__buffer = PACKET_START;
				}
				
				break;
		}
	}
	return succesfully_transmitted;
}

/*
 * Sending the message to superior class for processing
 * At the same time it can be cleaned from transceiver buffer
 * as the same message is never accessed twice
 */
std::string Transciever::get_message()
{
	std::string message;
	if(!__received_messages.empty())
	{
		message.assign(__received_messages.front());
		__received_messages.pop_front();
	}
	return message;
}

uint16_t Transciever::gencrc_16(const std::string & message)
{
	int length = message.size();
	uint8_t i;
	uint8_t p = 0;
	uint16_t crc = 0xffff;

	while (length--) {
		crc ^= message.at(p++) << 8;
		for (i=0; i < 8; i++)
			crc = crc & 0x8000 ? (crc << 1) ^ 0x1021 : crc << 1;
	}
	return crc & 0xffff;
}

int Transciever::process_received_data()
{
	
	uint16_t crc = 0;
	uint32_t msg_start, msg_end;
	std::string body;
	std::string crc_str;
	std::string count_str;

	msg_start = __buffer.find(BODY_DELIMITER);
	msg_end = __buffer.rfind(BODY_DELIMITER);

	// Transmitted data extracted from packet
	count_str = __buffer.substr(2,msg_start-2);
	body = __buffer.substr(msg_start + 1,msg_end-msg_start-1);
	crc_str = __buffer.substr(msg_end + 1, 4);
	
	// Removing data from buffer
	__buffer.clear();
	
	// Acknowledge
	if(body.find(ACK) != std::string::npos)
	{
		__received_messages.push_back(body);
		return 1;
	}

	try{
		crc = std::stoi(crc_str,nullptr,16);
	}catch(std::invalid_argument){
		return 0;
	}

#ifdef DEBUG_TRC
		std::cout << "got crc: " <<  crc_str << " should be: " << std::hex << gencrc_16(body) << "!" << std::endl;
		std::cout << "got counter: " << count_str << " should be: " << std::to_string(__counter_in) << std::endl;
		std::cout << "body: " << body << std::endl;
#endif

	// Check of CRC	
	if(gencrc_16(body) != crc)
		return 0;

#ifdef RECEIVER
	// Check of rolling counter
	if(count_str != std::to_string(__counter_in))
		return 0;
#endif
	__received_messages.push_back(body);
	__counter_in++;	
	return 1;
}

void Transciever::process_data_to_send(std::string & command, TRANSMITTION_PRIORITY priority)
{
	uint16_t crc = gencrc_16(command);

	std::stringstream ss;
	ss 	<< PACKET_START << std::to_string(__counter_out)
		<< BODY_DELIMITER << command
		<< BODY_DELIMITER << std::hex << crc << PACKET_END;

	// Message is put into output buffer based on its priority	
	if(priority == COMMAND)
		set_message(std::make_tuple( std::get<0>(__messages_to_send[0]) - 1 ,ss.str()), priority);	// To be sure that it gets removed 
	else
		set_message(std::make_tuple(__counter_out++,ss.str()), priority);
}

// Cleaning of the old messages from the buffer to be send
void Transciever::start_from_message(const std::string message)
{
	int counter;
	if(!message.empty())
	{
		try{
			counter = std::stoi(message,nullptr,10);
		}catch(std::invalid_argument){
			return;
		}
	}else
		counter = __counter_out;	
	
	for(size_t i = 0; i < __messages_to_send.size(); i++)
	{
#ifdef RECEIVER		
		if(std::get<1>(__messages_to_send[i]).find(SEND) !=std::string::npos)
		{
			if(std::get<0>(__messages_to_send[i]) >= counter)
				break;
			__messages_to_send.erase(__messages_to_send.begin());
			i--;
		}
#endif
#ifdef TRANSMITTER		
			if(std::get<0>(__messages_to_send[i]) == counter)
				break;
			__messages_to_send.erase(__messages_to_send.begin());
			i--;	
#endif			
	}
}
