#include "../include/transmitter.h"

// Generation of the name of the log file
std::string Transmitter::construct_name()
{
	time_t rawtime;
	struct tm * timeinfo;
	char buffer[80];

	time (&rawtime);
	timeinfo = localtime(&rawtime);

	strftime(buffer,sizeof(buffer),"%d-%m-%Y %H:%M:%S",timeinfo);
	std::string str(buffer);
	return str;
}

/*
 * The first is the initialization of the library wiringPi
 * then the iterative search for devices is performed
 * Then we initialize the log file and last we set size
 * of vectors used for getting the log messages and messages to 
 * transfer according to amount of connected devices
 */
int Transmitter::init_devices()
{
	// Setups the library
	if(wiringPiSetup() == -1)
	{
		__error_message = std::strerror (errno);
		return ERROR_INIT_FAILED;
	}

#ifdef DEBUG
	std::cout << "wiringPI initialized" << std::endl;
#endif

	__input_devices.clear();

	Device* device;
	DIR* dir = opendir(DEV_FOLDER.c_str());
	if(!dir)
	{
		__error_message = "Unable to open dev folder";
		return ERROR_INIT_FAILED;
	}

#ifdef DEBUG
	std::cout << "Opening the dev folder" << std::endl;
#endif

	struct dirent *ent;
	std::string filename;

	// Iterative search for devices
	while((ent = readdir(dir)) != NULL)
	{
		device = nullptr;
		filename = DEV_FOLDER + ent->d_name;
		if(filename.find(TTY_ACM) != std::string::npos)
			device = new GPS();
		else if(filename.find(TTY_USB) != std::string::npos)
			device = new Magnetometer();
		else if(filename.find(TTY_SERIAL0) != std::string::npos)
		{
			__transciever = new Transciever();
			if(__transciever->init_device(filename) != ERROR_SUCCESS)
				return ERROR_INIT_FAILED;
		}

		if(device && device->init_device(filename) == ERROR_SUCCESS)
			 __input_devices.push_front(device);
	}

#ifdef DEBUG
	std::cout << "closing the dev folder" << std::endl;
#endif

	// Initialization of logger
	__logger = new Logger();
	filename = 	"log/" + construct_name() + ".txt";
#ifdef DEBUG
	std::cout << "Generating file: " << filename << std::endl;
#endif
	if(__logger->init_device(filename) != ERROR_SUCCESS)
		return ERROR_FAILURE;

#ifdef DEBUG
	std::cout << "initialized logging file" << std::endl;
	std::cout << "Initialization finished. Recognized: "
			  << __input_devices.size() << " device(s)" << std::endl;
#endif

	// Setting size of the vectors for logging and transmitting		  
	__to_log.resize(__input_devices.size());
	__to_transfer.resize(__input_devices.size());

	return (__input_devices.size() > 0) ? ERROR_SUCCESS: ERROR_INIT_FAILED;
}

int Transmitter::run_loop()
{
	std::cout << "TRANSCIEVER SUCCESSFULLY INITIALIZED" << std::endl;
	__state = TRANSMITTER_READY;
	bool operating = true;
	bool transfer = false;
	bool updated = false;
	std::string tmp;
	std::string counter;

	while(operating)
	{		
		if((__transciever->data_available() &&
				__transciever->receive_data()) ||
					__transciever->number_of_messages_available() > 0)
						__state =  static_cast<TRANSMITTER_STATE>(process_received_data());		
		
		
		switch(__state)
		{
			case TRANSMITTER_LOG:
				for(size_t i = 0; i < __input_devices.size(); i++)
				{
					if(__input_devices.at(i)->data_available() &&
							__input_devices.at(i)->receive_data() &&
								__input_devices.at(i)->process_received_data() == ERROR_SUCCESS)
						{
							std::this_thread::sleep_for (std::chrono::microseconds(1));
							__to_log.at(i) = __input_devices.at(i)->get_to_log();
							__to_transfer.at(i) = __input_devices.at(i)->get_to_transfer();
							updated = true;
							
							//GPS is the first device
							if(i == 0)	
								transfer = true;
						}
				}
				
				if(updated)
				{
					__logger->log_data(__to_log);
					updated = false;
				}
				if(transfer)	
				{
					process_data_to_send(__to_transfer, DATA);		
					transfer = false;
				}
				break;
				
			case TRANSMITTER_TERMINATE:
				// Rather to be sure that the acknowledge was sent when program ends
				process_data_to_send({ACK},COMMAND);
				__transciever->send_data();

				terminate_connection();
				operating = false;
				break;
				
			case TRANSMITTER_RESET:
				terminate_connection();
				init_devices();
				__state = TRANSMITTER_READY;
				break;
				
			case TRANSMITTER_SEND:
				counter = __last_received.substr(__last_received.rfind(" ") + 1);
				__transciever->start_from_message(counter);
				__transciever->send_data();
				__state = TRANSMITTER_LOG;
				break;	
				
			default: 
				break;
		}
	}

	return ERROR_SUCCESS;
}

void Transmitter::process_data_to_send(std::vector<std::string> data, TRANSMITTION_PRIORITY priority)
{
	std::string tmp;
	tmp = accumulate(begin(data), end(data), tmp);
	__transciever->process_data_to_send(tmp, priority);
}

int Transmitter::process_received_data()
{
	__last_received = __transciever->get_message();
	// need of confirmation message
	if(__last_received.find(SEND) == std::string::npos) 
	{
		process_data_to_send({ACK},COMMAND);
		__transciever->send_data();
	}

	if(__last_received.find(LOG) != std::string::npos)
		return TRANSMITTER_LOG;
	if(__last_received.find(STOP) != std::string::npos)
		return TRANSMITTER_TERMINATE;
	if(__last_received.find(RESTART) != std::string::npos)
		return TRANSMITTER_RESET;
	if(__last_received.find(SEND) != std::string::npos)
		return TRANSMITTER_SEND;
	return __state;
}

void Transmitter::terminate_connection()
{
	for(auto device: __input_devices)
		device->terminate_connection();

	__logger->terminate_connection();
	__transciever->terminate_connection();
}
