from matplotlib.figure import Figure
from mpl_toolkits.axes_grid1.axes_divider import make_axes_locatable

from kivy.garden.matplotlib.backend_kivyagg import FigureCanvasKivyAgg
from kivy.clock import Clock
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.gridlayout import GridLayout
from kivy.uix.label import Label
from kivy.uix.screenmanager import Screen
from kivy.uix.button import Button
from kivy.uix.togglebutton import ToggleButton
from kivy.uix.widget import Widget
from kivy.graphics import Rectangle, Color

from threading import Thread, Lock
from multiprocessing import Process, Pipe

from data_gatherers import FileDataGatherer as DataGatherer  # FileDataGatherer, FifoDataGatherer, TestDataGatherer
from processing import DataProcessing
from interactive_plot import InteractivePlotter

import logging


class MainScreen(Screen):
    def __init__(self, interp_params, meas_params, general_params, **kwargs):
        super().__init__(**kwargs)

        meas_mode = meas_params["type"]  # The graph title must show either magnetic gradient or magnetic field
        if meas_mode == "vertical_gradient":
            self.graph = Graph(graph_title="Magnetic gradient (nT/m)")
        elif meas_mode == "single_sensor":
            self.graph = Graph(graph_title="Magnetic field (nT)")
        else:  # Should be unreachable if properly designed
            self.graph = Graph(graph_title="Unspecified")

        self.gatherer_pipe, ext_gatherer_pipe = Pipe()  # Local and external pipe ends are created, local is stored
        self.processing_pipe, ext_processing_pipe = Pipe()

        # Control UI components that will require updating ----------
        self.battery = 0
        self.battery_label = Label(text=str(self.battery)+"V", size_hint_x=0.3)
        self.packet_counter = 0
        self.counter_label = Label(text=str(self.packet_counter))
        # --------------------------------
        self.ack_popup = CustomPopup(msg="Command received.", pos_hint={"center_x": 0.5, "center_y": 0.5})
        self.error_popup = CustomPopup(msg="Error.", pos_hint={"center_x": 0.5, "center_y": 0.5})
        self.showing_popup = False  # Prevents stacking popups when multiple ACKs are received in quick succession
        self.showing_error = False
        self.ack_trigger = Clock.create_trigger(self.show_ack_popup)  # Used by any thread to show popup in the main one
        self.error_trigger = Clock.create_trigger(self.show_error_popup)

        # Data Gatherer and Processing processes launched, with threads in the main process as listeners ---
        gatherer_process = Process(target=DataGatherer,
                                   args=(ext_gatherer_pipe,),
                                   daemon=True)
        processing_process = Process(target=DataProcessing,
                                     args=(ext_processing_pipe, interp_params, meas_params, general_params.pop("step")),
                                     daemon=True)
        gatherer_process.start()
        processing_process.start()

        # Threads for handling concurrent new samples and processed data flows respectively
        Thread(target=self._parallel_gatherer_interface, daemon=True).start()
        Thread(target=self._parallel_processing_interface, daemon=True).start()
        # --------------------------------

        linear_only = (interp_params["type"] == "linear_only")  # Describes if custom interpolation is selected

        self.generate_screen_elements(linear_only)

    def generate_screen_elements(self, linear_only):
        """
        Defines how the graphical layout is built and structured
        ;param linear_only; Boolean. If true, only linear interpolation is available and no toggle button is needed
        """
        layout = BoxLayout(orientation="horizontal", spacing=10)
        inner_layout = BoxLayout(orientation="vertical", size_hint=(.2, 1), padding=[10, 10], spacing=6)

        # Gatherer controls ---------------
        gatherer_layout = BoxLayout(orientation="horizontal", padding=[0, 10], spacing=1, size_hint_y=0.5)

        gather_start_btn = Button(text="[size=11]START[/size]", markup=True)
        gather_start_btn.cmd = "log"
        gather_start_btn.bind(on_press=self._send_gatherer_cmd)
        gatherer_layout.add_widget(gather_start_btn)

        gather_pause_btn = Button(text="[size=11]PAUSE[/size]", markup=True)
        gather_pause_btn.cmd = "pause"
        gather_pause_btn.bind(on_press=self._send_gatherer_cmd)
        gatherer_layout.add_widget(gather_pause_btn)

        gather_end_btn = Button(text="[size=11]END[/size]", markup=True)
        gather_end_btn.cmd = "stop"
        gather_end_btn.bind(on_press=self._send_gatherer_cmd)
        gatherer_layout.add_widget(gather_end_btn)
        # End of gatherer controls ------

        inner_layout.add_widget(gatherer_layout)

        plot_pause_btn = ToggleButton(text="Pause plotting")
        plot_pause_btn.bind(on_press=self.graph.toggle_updating_figure)
        inner_layout.add_widget(plot_pause_btn)

        trace_toggle_btn = ToggleButton(text="Toggle trace", state="down")
        trace_toggle_btn.bind(on_press=self.graph.toggle_showing_trace)
        inner_layout.add_widget(trace_toggle_btn)

        if linear_only:  # No need for tricontour toggling, by default only plotting tricontours
            self.graph._tricontour_mode = True
        else:  # Tricontour toggle button is created
            tricontour_toggle_btn = ToggleButton(text="Toggle linear \n interpolation")
            tricontour_toggle_btn.bind(on_press=self.graph.toggle_tricontour_use)
            inner_layout.add_widget(tricontour_toggle_btn)

        external_plot_btn = Button(text="Interactive plot")
        external_plot_btn.bind(on_press=self.graph.interactive_plot)
        inner_layout.add_widget(external_plot_btn)

        # ----- Bottom control information ------
        control_layout = GridLayout(cols=2, padding=(10, 0), size_hint_y=None, spacing=[10, 0])

        counter_desc = Label(text="Samples: ")
        control_layout.add_widget(counter_desc)
        self.counter_label.size_hint_x = 0.3
        control_layout.add_widget(self.counter_label)

        control_layout.add_widget(self.battery_label)

        control_layout.height = len(control_layout.children)*15  # Forced height for proper display of all info

        inner_layout.add_widget(control_layout)
        # ----- End of bottom control -----

        layout.add_widget(inner_layout)
        layout.add_widget(self.graph.canvas)

        Clock.schedule_interval(self._update_control_info, 1)

        self.add_widget(layout)

    def _send_gatherer_cmd(self, instance, *args):
        """
        To be called through the gatherer control buttons.
        Sends the command to the gatherer associated with said button.
        :param instance: button that called this method
        """
        logging.info("[Main] Sending command - " + instance.cmd)
        try:
            self.gatherer_pipe.send(instance.cmd)
        except BrokenPipeError:
            self.error_trigger()

    def _parallel_gatherer_interface(self):
        """
        To be called on a different Thread.
        Handles the connection from the DataGatherer and towards the Processing.
        Checks each data packet, pops information not relevant to the processing side, and forwards the rest
        """

        packet = self.gatherer_pipe.recv()  # Initial non-empty packet to enter the loop

        while packet:  # Will exit if None or {} is received
            packet_type = packet.pop("type")
            if packet_type == "sample":
                self.battery = packet.pop("battery")
                self.processing_pipe.send(packet)
                self.packet_counter += 1

            elif packet_type == "ACK":
                self.ack_trigger()  # ACK Popup is shown

            packet = self.gatherer_pipe.recv()

        self.processing_pipe.send(None)  # None as a poison pill

    def _parallel_processing_interface(self):
        """
        To be called on a different Thread
        Handles the connection from the Processing.
        Calls the corresponding update method each time new data is received.
        """
        packet = self.processing_pipe.recv()

        while packet:
            packet_type = packet.pop("type")
            if packet_type == "raw":  # Length 3 for new trace plots (x_coord, y_coord, field_vect)
                self.graph.new_sample(packet["long"], packet["lat"], packet["data"])
            elif packet_type == "interp":
                self.graph.new_interpolated(packet["long"], packet["lat"], packet["data"])

            packet = self.processing_pipe.recv()

    def _update_control_info(self, *args):
        """
        Redraws the control information (sample counter and battery voltage).
        To be called in main thread.
        """
        self.counter_label.text = str(self.packet_counter)
        self.battery_label.text = str(self.battery) + "V"

    def show_ack_popup(self, *args):
        """
        Draws the ACK popup and schedules it to be deleted in 1 second.
        To prevent stacking of multiple popups when called in quick succession, it only draws it if not already drawn.
        To be called from the main thread or through ack_trigger.
        """
        if not self.showing_popup:
            self.showing_popup = True
            self.add_widget(self.ack_popup)
            Clock.schedule_once(self.remove_ack_popup, 1)

    def remove_ack_popup(self, *args):
        """
        Removes the ACK popup from the screen and sets the corresponding check variable to False.
        """
        self.remove_widget(self.ack_popup)
        self.showing_popup = False

    def show_error_popup(self, *args):
        """
        Draws the error popup and schedules it to be deleted in 1 second.
        To prevent stacking of multiple popups when called in quick succession, it only draws it if not already drawn.
        To be called from the main thread or through error_trigger.
        """
        if not self.showing_error:
            self.showing_error = True
            self.add_widget(self.error_popup)
            Clock.schedule_once(self.remove_error_popup, 1)

    def remove_error_popup(self, *args):
        """
        Removes the error popup from the screen and sets the corresponding check variable to False.
        """
        self.remove_widget(self.error_popup)
        self.showing_error = False


class CustomPopup(Widget):
    def __init__(self, msg, **kwargs):
        super().__init__(**kwargs)
        self.size_hint = (None, None)
        self.size = (200, 75)

        # Background is set as a rectangle
        with self.canvas:
            Color(.7, .15, .15, 1)
            self.background = Rectangle(pos=self.center, size=(self.width, self.height))

        self.label = Label(text=msg, pos=self.center, size=(self.width, self.height))
        self.add_widget(self.label)

        # Dynamic binding necessary to update rectangle and label size/pos when parent size/pos changes
        self.bind(pos=self.redraw, size=self.redraw)

    def redraw(self, *args):
        """
        Updates children position and size whenever the parent's position or size changes.
        Alternative to .kv language.
        """
        self.background.size = self.size
        self.background.pos = self.pos
        self.label.size = self.size
        self.label.pos = self.pos


class Graph:
    """
    Represents a contour (heatmap) + trace plot.
    Keeps track of both plots themselves as well as direct references to the information they show.
    Implements all the necessary logic to toggle between the available plotting modes, and limits the number
    of plot attempts through triggers.
    """

    def __init__(self, graph_title):
        """
        Sets the axes and internal variables necessary for operation.
        :param graph_title: Graph title may vary depending on interpolation mode, so is left as a variable
        """
        self.figure = Figure()
        self.canvas = FigureCanvasKivyAgg(self.figure)
        self.ax = self.figure.add_subplot(111)
        self.color_ax = make_axes_locatable(self.ax).append_axes("right", size="5%", pad="2%")  # For colorbar

        self.figure.suptitle(graph_title)

        # Triggers used to notify main thread of new sample/interpolated values
        self.update_xyz_trigger = Clock.create_trigger(self._update_xyz_plot)
        self.update_contour_trigger = Clock.create_trigger(self._update_contour_plot)

        self.ax.set_xlabel("Longitude")
        self.ax.set_ylabel("Latitude")

        # xyz_coord variables store the currently plotted values. May differ from the new_ ones if plotting is paused
        self.xx = []
        self.yy = []
        self.zz = []
        self.x_coord = []  # xyz_coord store raw data values: GPS coordinates and field values
        self.y_coord = []
        self.z_coord = []

        # new_XXX variables store the most updated vector of values. Can be updated from any thread (thread-safe)
        self.new_x_coord = []
        self.new_y_coord = []
        self.new_z_coord = []
        self.new_xx = []  # xx, yy, zz: interpolated vectors (xx/yy) and matrix (zz) of values
        self.new_yy = []
        self.new_zz = []

        # Locks to handle multithreaded access to new_XX and new_X_coord variables
        self.coord_lock = Lock()
        self.contour_lock = Lock()

        self.flight_plot, = self.ax.plot(self.x_coord, self.y_coord, 'kx-', markersize=1.5, linewidth=0.5)
        self.contour_plot = None

        # Operating configuration
        self._showing_trace = True  # Whether or not the flight trace is plotted at all
        self._showing_contour = True  # Whether or not the heatmap is plotted at all
        self._tricontour_mode = False  # (In case heatmap is plotted) whether it's a tricontour or normal contour
        self._updating_figure = True  # For pausing graph updates

        self.canvas.draw()

    def new_interpolated(self, xx, yy, zz):
        """
        THREAD SAFE
        Updates the latest stored interpolation data.
        Calls the plot redraw trigger to handle the new data as appropriate in main thread.
        """
        with self.contour_lock:
            self.new_xx = xx
            self.new_yy = yy
            self.new_zz = zz
        self.update_contour_trigger()

    def new_sample(self, new_x, new_y, new_z):
        """
        THREAD SAFE
        Takes three (1D) arrays and updates the latest stored coordinate data.
        Calls the plot redraw trigger to update the graph.
        """
        with self.coord_lock:
            self.new_x_coord = new_x
            self.new_y_coord = new_y
            self.new_z_coord = new_z

        self.update_xyz_trigger()  # The trigger is called to execute _handle_new_sample() in main thread

    def _update_xyz_plot(self, *args):
        """
        ONLY CALLED INTERNALLY THROUGH update_xyz_trigger.
        Executes on main thread.
        Checks whether or not to update existing plotted values.
        Executed only once in the next Kivy cycle, even if called multiple times (if called through trigger)

        :param args: This method is called by an Event trigger, which sends additional parameters.
        """
        if self._updating_figure:
            self._update_stored_coord()
            self._plot_redraw()

    def _update_contour_plot(self, *args):
        """
        ONLY CALLED INTERNALLY THROUGH update_contour_trigger.
        Executes on main thread.
        Checks whether or not to update existing plotted contour values.
        Executed only once in the next Kivy cycle, even if called multiple times (if called through trigger)

        :param args: This method is called by an Event trigger, which sends additional parameters.
        """
        if self._updating_figure:
            self._update_stored_contour()
            self._contour_redraw()

    def _update_stored_coord(self):
        """
        Updates stored (old) values for raw sample data values with the latest ones
        """
        with self.coord_lock:
            self.x_coord = self.new_x_coord
            self.y_coord = self.new_y_coord
            self.z_coord = self.new_z_coord

    def _update_stored_contour(self):
        """
        Updates stored (old) values for interpolated vectors and matrix with the latest ones
        """
        with self.contour_lock:
            self.xx = self.new_xx
            self.yy = self.new_yy
            self.zz = self.new_zz

    def toggle_tricontour_use(self, *args):
        """
        Used to toggle between tricontour and non-tricontour plotting.

        :param args: This method is called by a button, which sends additional parameters.
        """
        logging.info("Toggling tricontour")
        if self._tricontour_mode:
            # When tricontour plotting turned off, the contour is erased and contour_plot goes back to None
            self._tricontour_mode = False
            self._remove_contour()
            if self._showing_contour:  # If tricontour is turned off and heatmap is being shown, plot normal contour
                self._contour_redraw()
        else:
            self._tricontour_mode = True

        self._plot_redraw()  # Graph redrawn after toggling, no need to wait for next sample for an update

    def toggle_updating_figure(self, *args):
        """
        Toggles internal parameter _updating_figure, and updates the figure if turned on

        :param args: This method is called by a button, which sends additional parameters.
        """
        if self._updating_figure:
            self._updating_figure = False
        else:
            self._updating_figure = True
            self.update_xyz_trigger()
            self.update_contour_trigger()

    def toggle_showing_trace(self, *args):
        """
        Toggles internal parameter _showing_trace, and updates the figure accordingly.
        If it is turned off, the flight plot must be manually modified to not represent anything.

        :param args: This method is called by a button, which sends additional parameters.
        """
        if self._showing_trace:
            self.flight_plot.set_data([], [])
            self._showing_trace = False
        else:
            self._showing_trace = True

        self._plot_redraw()

    def _remove_contour(self):
        """
        Removes all contour components from the graph, since the whole contour cannot be removed in one single
            action, like a normal plot would.
        When a contour is removed, the stored value for it goes back to None. This allows the prevention of removing
            a not-existing contour (previously removed)
        """
        if self.contour_plot is not None:
            for col in self.contour_plot.collections:
                col.remove()
            self.color_ax.clear()
            self.contour_plot = None

    def _plot_redraw(self):
        """
        Updates all (selected) plots that depend on the stored raw data vectors (x_coord, y_coord, z_coord).
        Namely, updates the flight trace and tricontour plots (if possible).
        """
        if self._showing_trace:
            self.flight_plot.set_data(self.x_coord, self.y_coord)

        if self._showing_contour and self._tricontour_mode:
            if len(self.z_coord) > 4:
                self._remove_contour()
                self.contour_plot = self.ax.tricontourf(self.x_coord, self.y_coord, self.z_coord, levels=30)
                self.figure.colorbar(self.contour_plot, cax=self.color_ax)

        self.canvas.draw()

    def _contour_redraw(self):
        """
        Checks and redraws the heatmap if appropriate, removing the previous one.
        Only applicable for standard interpolation plots (not linear tricontour, which is done in _plot_redraw)
        Updates the drawn figure.
        """
        if self._showing_contour and not self._tricontour_mode:
            self._remove_contour()
            if len(self.zz) > 4:
                self.contour_plot = self.ax.contourf(self.xx, self.yy, self.zz, levels=100)
                self.figure.colorbar(self.contour_plot, cax=self.color_ax)

        self.canvas.draw()

    def interactive_plot(self, *args):
        """
        Gives current plot conditions and parameters to an external process which launches the default matplotlib
        interactive plot. Ensures no multithreading inconsistencies on given parameters through locks.
        """
        with self.coord_lock, self.contour_lock:
            if len(self.x_coord) < 4:
                return  # Prevents plotting attempts when very few samples are still available
            trace_data = (self.x_coord, self.y_coord)

            if self._tricontour_mode:
                contour_data = self.z_coord
            elif len(self.zz) < 4:
                return  # Prevents plotting attempts when interpolated matrix is not ready
            else:
                contour_data = (self.xx, self.yy, self.zz)

            p = Process(target=InteractivePlotter,
                        args=(self._showing_trace, trace_data, self._tricontour_mode, contour_data),
                        daemon=True)
            p.start()
