# -*- coding: utf-8 -*-

import numpy as np
from threading import Lock
import time  # For efficiency testing


class DataProcessing:
    """
    Processing core. Takes in new data samples, outputs computed data in the form of ready-to-plot 2-D matrices
    (for the heatmap) and vectors containing the updated flight trace.
    """
    def __init__(self, pipe, interp_params, meas_params, plot_step):
        """
        :param pipe: Connects with the application core. Used for receiving samples and sending computed data
        :param interp_params: Parameters strictly for interpolator initialization
        :param meas_params: Form of measuring the field (e.g. v_gradient, h_gradient, direct...)
        :param plot_step: Every how many samples the processed data is sent
        """
        self._SEND_PACKET_INTERVAL = plot_step  # # of processed packets between each plot update
        self.unsent_packets = 0  # Keeps track of the interval counter - send 1 out of X valid packets
        self.processing_pipe = pipe
        self.pipe_lock = Lock()
        self.raw_data = RawDataManager()  # Where sample field/coord values are stored/handled

        self.interp = None

        meas_type = meas_params["type"]

        if meas_type == "vertical_gradient":  # field_calc refers to the appropriate field calculation method
            self.V_SPACING = meas_params["spacing"]
            self.field_calc = self._v_grad_calc
        elif meas_type == "single_sensor":
            self.field_calc = self._single_magn_calc
        else:
            raise UnsupportedMeasModeException

        self.packet_collection(interp_params)

    def packet_collection(self, interp_params):
        """
        Main Data Processing loop.
        Collects incoming samples from the app_core.
        Checks integrity bit, and passes the previous one to the interpolator and vector builder.
        Packets are processed if they and the ones immediately before/after didn't have integrity problems.
        """

        # First packet is received for initial point, in case it's necessary for initializing interpolator
        prev_packet = self.processing_pipe.recv()
        interp_params["point_zero"] = (prev_packet["longitude"], prev_packet["latitude"])

        interp_type = interp_params.pop("type")
        if interp_type == "inv_dist_squared":
            self.interp = InverseDistanceInterpolator(pipe=self.processing_pipe, **interp_params)
        elif interp_type == "linear_only":
            self.interp = DummyInterpolator(pipe=self.processing_pipe, **interp_params)
        else:
            raise UnsupportedInterpolationException

        # There should be many packets in a survey - "None" in the first two not considered

        current_packet = self.processing_pipe.recv()

        prev_integrity = prev_packet.pop("integrity_lock")
        current_integrity = current_packet.pop("integrity_lock")

        if prev_integrity and current_integrity:  # First packet needs special treatment - not surrounded by other 2
            self.process_packet(prev_packet)

        next_packet = self.processing_pipe.recv()

        while next_packet:  # Breaks when None is received
            next_integrity = next_packet.pop("integrity_lock")
            # Packet is processed if both it and its immediate neighbours have proper integrity lock
            if prev_integrity and current_integrity and next_integrity:
                self.process_packet(current_packet)
            else:
                # If one of the three has problems, integrity variables must cycle (not all True)
                prev_integrity, current_integrity = current_integrity, next_integrity
            # No need to keep track of prev_packet (already sent/discarded), only of its integrity state
            current_packet = next_packet
            next_packet = self.processing_pipe.recv()

        # Last packet before the None also needs special treatment
        self.unsent_packets = self._SEND_PACKET_INTERVAL  # Hack to force sending the last samples
        if prev_integrity and current_integrity:
            self.process_packet(current_packet)

        # Once a None packet is received, it is sent back as poison pill for the _parallel_processing_interface() thread
        self.send(None)

    def process_packet(self, packet):
        """
        Performs all necessary operations to a valid packet:
        "Field" value calculation, appending to raw data vectors, feeding to interpolator and sending if appropriate.
        :param packet: Input, pre-processed packet (dictionary)
        """
        self.field_calc(packet)
        self.raw_data.new_sample(packet)
        self.interp.new_sample(packet["longitude"], packet["latitude"], packet["field"])
        self.unsent_packets += 1
        if self.unsent_packets == self._SEND_PACKET_INTERVAL:
            self.unsent_packets = 0
            self.send(self.raw_data.get_stored_data())
            self.interp.send()

    def send(self, data_vect):
        """
        Sends stored sample data vectors
        A lock is needed to account for interpolators that do multithreaded use of this pipe.
        :param data_vect: Packet containing accumulated sample values: {x_vect, y_vect, z_vect}
        """
        with self.pipe_lock:
            self.processing_pipe.send(data_vect)

    def _v_grad_calc(self, packet):
        """
        To be used through the field_calc method if necessary.
        Modifies the packet to take two different magnetic field readings and translate them to a single "field" value
        that represents the vertical gradient of the field.
        :param packet: Packet containing lower and upper magnetometer readings
        """
        packet["field"] = -(packet.pop("reading_2") - packet.pop("reading_1")) / self.V_SPACING

    def _single_magn_calc(self, packet):
        """
        To be used through the field_calc method if necessary.
        Modifies the packet to change the "reading_l" value for a "field" value.
        Trivial method, used to have a shared API with more complex, multi-sensor field calculations.
        :param packet: Packet containing lower and upper magnetometer readings
        """
        packet["field"] = packet.pop("reading_1")


class RawDataManager:
    """
    Receives new single samples and efficiently puts them into an array. Provides a simple API for accessing
    such an array where normally only a part of it contains valuable information. This way there is no need
    to take indexing into consideration when the stored useful data is to be accessed.
    """
    def __init__(self):
        self.capacity = 200
        self.elements = 0
        self.longitude_vect = np.zeros(self.capacity)
        self.latitude_vect = np.zeros(self.capacity)
        self.field_vect = np.zeros(self.capacity)

    def new_sample(self, packet):
        """
        Takes in packets and efficiently adds their value to a dynamic array
        :param packet: pre-processed packet where magnetometer readings have been translated to a field value
        """
        if self.elements == self.capacity:  # Buffer resizing in case limit is reached
            self.capacity *= 2

            new_longitude_vect = np.zeros(self.capacity)
            new_latitude_vect = np.zeros(self.capacity)
            new_field_vect = np.zeros(self.capacity)

            new_longitude_vect[:self.elements] = self.longitude_vect
            new_latitude_vect[:self.elements] = self.latitude_vect
            new_field_vect[:self.elements] = self.field_vect

            self.longitude_vect = new_longitude_vect
            self.latitude_vect = new_latitude_vect
            self.field_vect = new_field_vect

        self.longitude_vect[self.elements] = packet["longitude"]
        self.latitude_vect[self.elements] = packet["latitude"]
        self.field_vect[self.elements] = packet["field"]
        self.elements += 1

    def get_stored_data(self):
        """
        Returns a tuple containing all samples so far
        :return: (longitude, latitude, field) where each is a vector with an element per sample
        """
        return {"type": "raw",
                "long": self.longitude_vect[:self.elements],
                "lat": self.latitude_vect[:self.elements],
                "data": self.field_vect[:self.elements]}


class InverseDistanceInterpolator:
    def __init__(self, pipe, point_zero, step_meters=0.1, dmin=0.01, dmax=8):
        """
        :param pipe: Pipe through which the interpolated result is sent
        :param step_meters: Grid cell spacing
        :param point_zero: Initial point for the interpolation: (longitude, latitude), due to similarity with (x, y)
        notation, in decimal degrees
        """

        self.pipe_out = pipe

        # Constants for distance calculation ---------
        earth_radius = 6371000
        expand_meters = 15  # Distance (m) to be added to one side when the matrix is expanded

        self._D_MIN = dmin  # Max/min distance for weight function (meters)
        self._D_MAX = dmax
        self._WINDOW_SIZE = int(2*self._D_MAX/step_meters + 2)

        self._LAT_STEP = 180/np.pi*step_meters/earth_radius  # Separation in latitude between contiguous grid points
        self._LONG_STEP = self._LAT_STEP/np.cos(point_zero[1]*np.pi/180)
        self._EXPAND_SIZE = int(expand_meters/step_meters)  # number of rows/columns to be added each time

        # Auxiliary for boundary check: Prevent new samples being within current range but not whole window around them
        self._D_MAX_LAT = (self._D_MAX/step_meters + 2)*self._LAT_STEP  # Note: +2 to have 2 extra _LAT_STEP of margin
        self._D_MAX_LONG = (self._D_MAX/step_meters + 2)*self._LONG_STEP

        self.interp_matrix = np.zeros((2*self._EXPAND_SIZE + 1, 2*self._EXPAND_SIZE + 1))  # Zero-filled¡
        self.accumulated_D_matrix = np.zeros((2*self._EXPAND_SIZE + 1, 2*self._EXPAND_SIZE + 1))

        # interp_long/lat, mesh_long/lat: stored for efficiency
        # np.arange does not include the last (stop) element. size+0.5 used to ensure it does include all
        self.interp_long = np.arange(start=(point_zero[0] - self._EXPAND_SIZE*self._LONG_STEP),
                                     stop=point_zero[0] + (self._EXPAND_SIZE+0.5)*self._LONG_STEP,
                                     step=self._LONG_STEP)
        self.interp_lat = np.arange(start=(point_zero[1] - self._EXPAND_SIZE*self._LAT_STEP),
                                    stop=point_zero[1] + (self._EXPAND_SIZE+0.5)*self._LAT_STEP,
                                    step=self._LAT_STEP)
        # Meshgrid from references to originals (no copying for speed)
        self.mesh_long, self.mesh_lat = np.meshgrid(self.interp_long, self.interp_lat, copy=False)

        (self.lat_min, self.lat_max) = (self.interp_lat[0], self.interp_lat[-1])
        (self.long_min, self.long_max) = (self.interp_long[0], self.interp_long[-1])
        """
        # Debugging ------------------
        print("Expand size (elements): "+str(self._EXPAND_SIZE)+" - Window size (elements): "+str(self._WINDOW_SIZE))
        print("LONG_STEP: "+str(self._LONG_STEP)+" - LAT_STEP: "+str(self._LAT_STEP))
        print("Initial interp_long: "+str(self.interp_long.shape)+" - Initial interp_lat: "+str(self.interp_lat.shape))
        print("Point_zero: "+str(point_zero))
        print("Long limits: "+str((self.long_min, self.long_max))+" - Lat limits: "+str((self.lat_min, self.lat_max)))
        """

    def new_sample(self, new_long, new_lat, new_field):
        """
        Takes a new sample with its longitude, latitude and (post-processed) field values, and performs an
        interpolation using the previous samples together with the current one. When it is done, it sends the
        result through the associated output pipe.
        :param new_long:
        :param new_lat:
        :param new_field:
        :return:
        """
        """
        # Debugging -------------------
        print("------------NEW SAMPLE------------------")
        print("New sample: "+str((new_long, new_lat)))
        # -------------------------------------------
        """
        self._check_boundaries(new_long, new_lat)

        # Indexes of moving window of points to update (up to, not including the stop)
        long_center_flindex = (new_long - self.long_min)/self._LONG_STEP  # Float approximation of longitude index
        long_start_index = int(long_center_flindex - self._WINDOW_SIZE/2)
        long_stop_index = long_start_index + self._WINDOW_SIZE
        lat_center_flindex = (new_lat - self.lat_min)/self._LAT_STEP
        lat_start_index = int(lat_center_flindex - self._WINDOW_SIZE/2)
        lat_stop_index = lat_start_index + self._WINDOW_SIZE

        # First latitude (row), then longitude (column). Counter-intuitive:
        windowed_interp = self.interp_matrix[lat_start_index: lat_stop_index, long_start_index: long_stop_index]
        windowed_D = self.accumulated_D_matrix[lat_start_index: lat_stop_index, long_start_index: long_stop_index]
        windowed_lat = self.mesh_lat[lat_start_index: lat_stop_index, long_start_index: long_stop_index]
        windowed_long = self.mesh_long[lat_start_index: lat_stop_index, long_start_index: long_stop_index]

        distances = haversine_distance(windowed_long, windowed_lat, new_long, new_lat)

        weighed_distances = self._weighting_function(distances)

        # Using [:, :] updates the values in the original interp_matrix instead of overwriting windowed_interp variable
        windowed_interp[:, :] *= windowed_D
        windowed_interp[:, :] += weighed_distances*new_field
        windowed_D[:, :] += weighed_distances
        aux_windowed_D = windowed_D > 0
        windowed_interp[aux_windowed_D] /= windowed_D[aux_windowed_D]

        """
        # For visualizing window sizes only
        self.interp_matrix *= 0
        windowed_interp[:, :] = distances
        """

    def _check_boundaries(self, long, lat):
        """
        Checks if the indicated longitude and latitude fall withing the existing grid, and enlarges the grid
        if they are outside.
        Grid is enlarged by a fixed amount (_EXPAND_SIZE rows/columns), and the new sample is assumed to fall inside.
        :param long: New sample's longitude (decimal degrees)
        :param lat: New sample's latitude (decimal degrees)
        """
        """
        # Debugging ------------------------------
        print("-----------CHECK BOUNDARIES--------------")
        print("Longitude limits: "+str((self.long_min, self.long_max))+\
            " - Latitude limits: "+str((self.lat_min, self.lat_max)))
        print("Size of interp_matrix before expansion: " + str(self.interp_matrix.shape))
        print("size of accumulated_D_matrix before expansion: " + str(self.accumulated_D_matrix.shape))
        # --------------------------------------------------------
        """
        if (long+self._D_MAX_LONG) > self.long_max:
            # A zero-filled matrix is appended "on the right" of interp_matrix. inter_long array regenerated
            # Number of appended columns: _EXPAND_SIZE + however many are necessary between long_max and the new long
            total_expand = self._EXPAND_SIZE + int((long+self._D_MAX_LONG - self.long_max)/self._LONG_STEP)
            self.long_max += total_expand*self._LONG_STEP
            appended_array1 = np.zeros((self.interp_matrix.shape[0], total_expand))
            appended_array2 = np.zeros((self.interp_matrix.shape[0], total_expand))

            self.interp_matrix = np.hstack((self.interp_matrix, appended_array1))
            self.accumulated_D_matrix = np.hstack((self.accumulated_D_matrix, appended_array2))
            self.interp_long = np.arange(start=self.long_min,
                                         stop=self.long_max + 0.5*self._LONG_STEP,  # +0.5 to avoid rounding errors
                                         step=self._LONG_STEP)
            # Meshgrids need to be re-created
            self.mesh_long, self.mesh_lat = np.meshgrid(self.interp_long, self.interp_lat, copy=False)

        elif (long-self._D_MAX_LONG) < self.long_min:
            # A zero-filled matrix is appended "on the left" of interp_matrix. interp_long array regenerated
            # Number of appended columns: _EXPAND_SIZE + however many are necessary between the new long and long_min
            total_expand = self._EXPAND_SIZE + int((self.long_min - (long-self._D_MAX_LONG))/self._LONG_STEP)
            self.long_min -= total_expand*self._LONG_STEP
            appended_array1 = np.zeros((self.interp_matrix.shape[0], total_expand))
            appended_array2 = np.zeros((self.interp_matrix.shape[0], total_expand))

            self.interp_matrix = np.hstack((appended_array1, self.interp_matrix))
            self.accumulated_D_matrix = np.hstack((appended_array2, self.accumulated_D_matrix))
            self.interp_long = np.arange(start=self.long_min,
                                         stop=self.long_max + 0.5*self._LONG_STEP,
                                         step=self._LONG_STEP)
            self.mesh_long, self.mesh_lat = np.meshgrid(self.interp_long, self.interp_lat, copy=False)

        if (lat+self._D_MAX_LAT) > self.lat_max:
            # A zero-filled matrix is appended at the end of interp_matrix. interp_lat array regenerated
            # Number of appended rows: _EXPAND_SIZE + however many are necessary between the new lat and lat_max
            total_expand = self._EXPAND_SIZE + int((lat+self._D_MAX_LAT - self.lat_max)/self._LAT_STEP)
            self.lat_max += total_expand*self._LAT_STEP
            appended_array1 = np.zeros((total_expand, self.interp_matrix.shape[1]))
            appended_array2 = np.zeros((total_expand, self.interp_matrix.shape[1]))

            self.interp_matrix = np.vstack((self.interp_matrix, appended_array1))
            self.accumulated_D_matrix = np.vstack((self.accumulated_D_matrix, appended_array2))

            self.interp_lat = np.arange(start=self.lat_min,
                                        stop=self.lat_max + 0.5*self._LAT_STEP,
                                        step=self._LAT_STEP)
            self.mesh_long, self.mesh_lat = np.meshgrid(self.interp_long, self.interp_lat, copy=False)

        elif (lat-self._D_MAX_LAT) < self.lat_min:
            # A zero-filled matrix is appended at the beginning of interp_matrix. interp_lat array regenerated
            # Number of appended rows: _EXPAND_SIZE + however many are necessary between lat_min and the new lat
            total_expand = self._EXPAND_SIZE + int((self.lat_min - (lat - self._D_MAX_LAT))/self._LAT_STEP)
            self.lat_min -= total_expand*self._LAT_STEP
            appended_array1 = np.zeros((total_expand, self.interp_matrix.shape[1]))
            appended_array2 = np.zeros((total_expand, self.interp_matrix.shape[1]))

            self.interp_matrix = np.vstack((appended_array1, self.interp_matrix))
            self.accumulated_D_matrix = np.vstack((appended_array2, self.accumulated_D_matrix))

            self.interp_lat = np.arange(start=self.lat_min,
                                        stop=self.lat_max + 0.5*self._LAT_STEP,
                                        step=self._LAT_STEP)
            self.mesh_long, self.mesh_lat = np.meshgrid(self.interp_long, self.interp_lat, copy=False)
        """
        # - Debugging ----------------------------------------------
        print("Size of interp_matrix after expansion: " + str(self.interp_matrix.shape))
        print("size of accumulated_D_matrix after expansion: " + str(self.accumulated_D_matrix.shape))
        print("New? Longitude limits: "+str((self.long_min, self.long_max))+\
            " - Latitude limits: "+str((self.lat_min, self.lat_max)))
        ("--------------------END OF CHECK BOUNDARIES------------------------")
        # print----------------------------------------------------------------------
        """

    def _weighting_function(self, dist_matrix):
        """
        Generates the result of the interpolator weighting function. Works on the whole distance matrix at once
        :param dist_matrix: matrix containing the distance to the newest sample from all points within the window
        :return: Associated weighting function result to the input distance (matrix)
        """
        dist_matrix[dist_matrix < self._D_MIN] = self._D_MIN
        dist_matrix[dist_matrix > self._D_MAX] = np.inf  # Numpy handles division by infinite, could be improved
        return 1/(dist_matrix*dist_matrix)

    def send(self):
        # Masking the interpolated matrix to hide non-updated values.
        masked_matrix = np.ma.array(self.interp_matrix, copy=False, mask=(self.interp_matrix == 0))
        dict_out = {"type": "interp", "long": self.interp_long, "lat": self.interp_lat, "data": masked_matrix}
        self.pipe_out.send(dict_out)


class DummyInterpolator:
    """
    Interpolator used for linear-only interpolation, which requires no interpolated matrix generation.
    To avoid unnecessary overhead, no processing is done and no interpolated data is sent back to the main program.
    This class only exists to provide a consistent interpolator-independent flow for the packet handling.
    """
    def __init__(self, pipe, **kwargs):
        pass

    def new_sample(self, *args, **kwargs):
        pass

    def send(self, *args, **kwargs):
        pass


def haversine_distance(long_a, lat_a, long_b, lat_b):
    """
    Calculates great-circle distance (in meters) between points A and B using the haversine formula (approximation).
    Errors below 0.6%.
    All parameters in degrees.
    Works for array inputs as long as there is the same number of points in A and B, or one of both is a single point.
    :param long_a: Longitude of point A.
    :param lat_a: Latitude of point A.
    :param long_b: Longitude of point B.
    :param lat_b: Latitude of point B.
    :return: Distance in meters between two points. If one (or both) inputs are an array, returns array of distances
    """
    earth_radius = 6371000
    to_radians = np.pi/180

    lat_a_rad = lat_a*to_radians
    lat_b_rad = lat_b*to_radians
    delta_long_rad = (long_b - long_a)*to_radians
    delta_lat_rad = lat_b_rad - lat_a_rad

    aux_lat = np.sin(delta_lat_rad/2)
    aux_long = np.sin(delta_long_rad/2)

    a = aux_lat*aux_lat + np.cos(lat_a_rad)*np.cos(lat_b_rad)*aux_long*aux_long
    c = 2*np.arctan2(np.sqrt(a), np.sqrt(1-a))
    return earth_radius*c


class UnsupportedMeasModeException(Exception):
    pass


class UnsupportedInterpolationException(Exception):
    pass
