from kivy.core.window import Window
from kivy.uix.textinput import TextInput
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.accordion import Accordion, AccordionItem
from kivy.uix.slider import Slider
from kivy.uix.filechooser import FileChooserIconView
from kivy.uix.label import Label
from kivy.uix.gridlayout import GridLayout
from kivy.uix.screenmanager import Screen
from kivy.uix.button import Button
from kivy.uix.widget import Widget

from main_screen import MainScreen


class StartScreen(Screen):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        Window.clearcolor = (.15, .15, .15, 1)  # Dark gray background

        layout = BoxLayout(orientation="vertical")

        inner_layout = GridLayout(cols=3)
        inner_layout.add_widget(Label(text="Interpolation", size_hint=(0.3, 0.09), font_size='20dp'))
        inner_layout.add_widget(Label(text="Survey type", size_hint=(0.3, 0.09), font_size='20dp'))
        inner_layout.add_widget(Label(text="General", size_hint=(0.3, 0.09), font_size='20dp'))

        self.interp_accordion = InterpolationAccordion()
        self.mode_accordion = MeasurementModeAccordion()
        self.general_settings = GeneralSettingsLayout()

        inner_layout.add_widget(self.interp_accordion)
        inner_layout.add_widget(self.mode_accordion)
        inner_layout.add_widget(self.general_settings)

        continue_btn = Button(text="START", size_hint=(1, 0.07))
        continue_btn.bind(on_press=self.proceed_to_main)

        layout.add_widget(inner_layout)
        layout.add_widget(continue_btn)

        self.add_widget(layout)

    def proceed_to_main(self, *args):
        interp_params = self.interp_accordion.selected_params
        meas_params = self.mode_accordion.selected_params
        general_params = self.general_settings.params

        self.manager.add_widget(MainScreen(interp_params, meas_params, general_params, name="main"))
        self.manager.current = "main"


# -------------------- Interpolation -----------------
class InterpolationAccordion(Accordion):
    """
    Contains all the interpolation methods in an accordion fashion.
    To add a new one, create a new AccordionItem with the corresponding fields, and a method to update.

    CAREFUL - Parameters to the interpolator (in processing.py) are passed with the params dictionary key
    as a variable name. Dictionary keys must be chosen accordingly.
    """
    def __init__(self, **kwargs):
        super().__init__(orientation="vertical", **kwargs)
        self._selected = None  # AccordionItem representing selected interpolation method

        lin = LinearOnlyItem()
        lin.bind(collapse=self.update_selected)
        self.add_widget(lin)

        id2 = InverseDistanceSquaredItem()
        id2.bind(collapse=self.update_selected)
        self.add_widget(id2)

    @property
    def selected_params(self):
        """
        Returns the dictionary containing the start parameters corresponding to the selected interpolation method.
        """
        return self._selected.params

    def update_selected(self, instance, collapse):
        """
        Keeps track of which on the available Accordion items is selected, so the parameters for the appropriate
        mode can be extracted when initializing MainScreen.
        :param instance: Accordion Item that called this method upon being updated
        :param collapse: Boolean. Represents updated state of the AccordionItem (False = selected)
        """
        if not collapse:  # To only have the newly uncollapsed (selected) accordion item
            self._selected = instance


class InverseDistanceSquaredItem(AccordionItem):
    """
    Accordion Item to be place inside the InterpolationAccordion. Includes all necessary fields,
    parameters and handling functions for an Inverse Distance Squared interpolator.
    """

    def __init__(self, **kwargs):
        super().__init__(title="Inverse Squared Distance", **kwargs)

        min_window_size = 2
        max_window_size = 15
        init_window_size = 6
        min_cell_size = 0.4
        max_cell_size = 2
        init_cell_size = 0.8
        self.params = {"type": "inv_dist_squared"}

        layout = BoxLayout(orientation="vertical", padding=(10, 10))

        desc = "-Window (meters): Radius of affected interpolated points around each sample\n\
-Cell size (meters): Separation between interpolated points around each sample."
        desc_label = Label(text=desc, size_hint=(1, None), padding=(0, 10))
        # Over-complicated workaround to avoid using .kv file:
        desc_label.bind(
            width=lambda *x: desc_label.setter('text_size')(desc_label, (desc_label.width, None)),
            texture_size=lambda *x: desc_label.setter('height')(desc_label, desc_label.texture_size[1]))

        layout_inner = GridLayout(cols=3)

        # Window size slider and labels
        layout_inner.add_widget(Label(text="Window", size_hint=(0.7, 1)))
        window_slider = Slider(min=min_window_size, max=max_window_size, size_hint=(2, 1))
        window_slider.bind(value=self.on_change_window)
        layout_inner.add_widget(window_slider)
        self.window_value = Label(size_hint=(0.3, 1))  # Need to keep reference for updating value label
        layout_inner.add_widget(self.window_value)

        window_slider.value = init_window_size

        # Cell size slider and labels
        layout_inner.add_widget(Label(text="Cell size", size_hint=(0.7, 1)))
        cell_slider = Slider(min=min_cell_size, max=max_cell_size, size_hint=(2, 1))
        cell_slider.bind(value=self.on_change_cell)
        layout_inner.add_widget(cell_slider)
        self.cell_value = Label(size_hint=(0.3, 1))  # Need to keep reference for updating value label
        layout_inner.add_widget(self.cell_value)

        cell_slider.value = init_cell_size

        layout.add_widget(desc_label)
        layout.add_widget(layout_inner)
        self.add_widget(layout)

    def on_change_window(self, instance, value):
        self.window_value.text = "{:04.1f}".format(value)
        self.params["dmax"] = value

    def on_change_cell(self, instance, value):
        self.cell_value.text = "{:04.2f}".format(value)
        self.params["step_meters"] = value


class LinearOnlyItem(AccordionItem):

    def __init__(self, **kwargs):
        super().__init__(title="Linear only", **kwargs)

        self.params = {"type": "linear_only"}

        layout = BoxLayout(orientation="vertical", padding=(10, 10))

        desc = "Linear Interpolation can be toggled in all other interpolation modes, \
but the main selected mode is kept running in the background.\n\n\
Selecting this mode avoids additional computation in the background for increased performance."
        desc_label = Label(text=desc, padding=(0, 10))
        # Over-complicated workaround to avoid using .kv file:
        desc_label.bind(
            width=lambda *x: desc_label.setter('text_size')(desc_label, (desc_label.width, None)),
            texture_size=lambda *x: desc_label.setter('height')(desc_label, desc_label.texture_size[1]))

        layout.add_widget(desc_label)
        self.add_widget(layout)


# ----------------------- Measurement mode -------------------
class MeasurementModeAccordion(Accordion):
    """
    -Contains all the measurement modes in an accordion fashion.
    -To add a new one, create a new AccordionItem with the corresponding fields, and a method to update each into
    the parameter dictionary. Bind that AccordionItem to the update_selected method like the existing ones.
    """
    def __init__(self, **kwargs):
        super().__init__(orientation="vertical", **kwargs)

        self._selected = None  # AccordionItem representing selected measurement mode

        single_sensor = SingleSensorItem()
        single_sensor.bind(collapse=self.update_selected)

        v_gradient = VerticalGradientItem()
        v_gradient.bind(collapse=self.update_selected)

        self.add_widget(single_sensor)
        self.add_widget(v_gradient)

    @property
    def selected_params(self):
        """
        Returns the dictionary containing the start parameters corresponding to the selected measurement mode.
        """
        return self._selected.params

    def update_selected(self, instance, collapse):
        """
        Keeps track of which on the available Accordion items is selected, so the parameters for the appropriate
        mode can be extracted when initializing MainScreen.
        :param instance: Accordion Item that called this method upon being updated
        :param collapse: Boolean. Represents updated state of the AccordionItem (False = selected)
        """
        if not collapse:
            self._selected = instance


class SingleSensorItem(AccordionItem):
    def __init__(self, **kwargs):
        super().__init__(title="Single sensor", **kwargs)
        self.params = {"type": "single_sensor"}

        layout = BoxLayout(orientation="vertical", padding=(10, 10))

        desc = "A single magnetometer is used, and the field values are drawn directly. This does not correct\
for time or spacial variations on the field, and those must be filtered."
        desc_label = Label(text=desc, size_hint=(1, None), padding=(0, 10))
        # Over-complicated workaround to avoid using .kv file:
        desc_label.bind(
            width=lambda *x: desc_label.setter('text_size')(desc_label, (desc_label.width, None)),
            texture_size=lambda *x: desc_label.setter('height')(desc_label, desc_label.texture_size[1]))

        separator_widget = Widget()

        layout.add_widget(desc_label)
        layout.add_widget(separator_widget)
        self.add_widget(layout)


class VerticalGradientItem(AccordionItem):
    def __init__(self, **kwargs):
        super().__init__(title="Vertical gradient", **kwargs)
        self.params = {"type": "vertical_gradient"}
        min_separation = 0.10
        max_separation = 3
        default_separation = 1

        layout = BoxLayout(orientation="vertical", padding=(10, 10))

        desc = "Two magnetometers are used to calculate the vertical gradient of the magnetic field at each point. \
This filters field time variations and other anomalies. The gradient will be directly plotted.\n\n \
-Separation (meters): Vertical distance between the two magnetometers."
        desc_label = Label(text=desc, size_hint=(1, None), padding=(0, 10))
        # Over-complicated workaround to avoid using .kv file:
        desc_label.bind(
            width=lambda *x: desc_label.setter('text_size')(desc_label, (desc_label.width, None)),
            texture_size=lambda *x: desc_label.setter('height')(desc_label, desc_label.texture_size[1]))

        layout_inner = GridLayout(cols=3)

        layout_inner.add_widget(Label(text="Separation", size_hint=(0.7, 1)))
        separation_slider = Slider(min=min_separation, max=max_separation, size_hint=(2, 1))
        separation_slider.bind(value=self.on_change_separation)
        layout_inner.add_widget(separation_slider)
        self.separation_value = Label(size_hint=(0.3, 1))  # Need to keep reference for updating
        layout_inner.add_widget(self.separation_value)

        separation_slider.value = default_separation
        layout.add_widget(desc_label)
        layout.add_widget(layout_inner)
        self.add_widget(layout)

    def on_change_separation(self, instance, value):
        self.separation_value.text = "{:05.2f}".format(value)
        self.params["spacing"] = value


# ----------------------- Initial position information ---------
class PositionAccordion(Accordion):
    """
    -Contains all the methods for entering a-priori location information.
    -To add a new one, create a new AccordionItem with the corresponding fields, and a method to update each into
    the parameter dictionary. Bind that AccordionItem to the update_selected method like the existing ones.
    """
    def __init__(self, **kwargs):
        super().__init__(orientation="vertical", **kwargs)

        self._selected = None  # AccordionItem representing selected a-priori location info

        # from_file = FromFileItem()
        # from_file.bind(collapse=self.update_selected)
        # self.add_widget(from_file)

        manual = ManualItem()
        manual.bind(collapse=self.update_selected)
        self.add_widget(manual)

        no_info = NoInfoItem()
        no_info.bind(collapse=self.update_selected)
        self.add_widget(no_info)

    @property
    def selected_params(self):
        """
        Returns the dictionary containing the start parameters corresponding to the selected a-priori position info.
        """
        return self._selected.params

    def update_selected(self, instance, collapse):
        """
        Keeps track of which on the available Accordion items is selected, so the parameters for the appropriate
        mode can be extracted when initializing MainScreen.
        :param instance: Accordion Item that called this method upon being updated
        :param collapse: Boolean. Represents updated state of the AccordionItem (False = selected)
        """
        if not collapse:
            self._selected = instance


class NoInfoItem(AccordionItem):
    def __init__(self, **kwargs):
        super().__init__(title="No a-priori info", **kwargs)

        self.params = {"type": "no_info"}

        self.add_widget(Label(text="First sample will be\nused as starting point"))


class ManualItem(AccordionItem):
    def __init__(self, **kwargs):
        super().__init__(title="Manual selection", **kwargs)

        self.params = {"type": "manual"}

        self.add_widget(TextInput(text="Latitude", multiline=False, size_hint=(1, 0.2)))
        self.add_widget(TextInput(text="Longitude", multiline=False, size_hint=(1, 0.2)))


class GeneralSettingsLayout(BoxLayout):
    def __init__(self, **kwargs):
        super().__init__(orientation="vertical", padding=(10, 10), **kwargs)
        self.params = {}
        desc = "-Plot step: number of processed samples for each plot update. Set high at low-end systems or high \
sample rates for better performance."
        desc_label = Label(text=desc, size_hint=(1, None), padding=(0, 10))
        # Over-complicated workaround to avoid using .kv file:
        desc_label.bind(
            width=lambda *x: desc_label.setter('text_size')(desc_label, (desc_label.width, None)),
            texture_size=lambda *x: desc_label.setter('height')(desc_label, desc_label.texture_size[1]))
        self.add_widget(desc_label)

        layout_inner = GridLayout(cols=3)

        # Plot step slider and labels
        layout_inner.add_widget(Label(text="Plot step", size_hint=(0.7, 1)))
        step_slider = Slider(min=1, max=10, size_hint=(2, 1), step=1)
        step_slider.bind(value=self.on_change_step)
        layout_inner.add_widget(step_slider)
        self.step_value = Label(size_hint=(0.3, 1))  # Need to keep reference for updating
        layout_inner.add_widget(self.step_value)
        step_slider.value = 2

        self.add_widget(layout_inner)

    def on_change_step(self, instance, value):
        self.step_value.text = str(value)
        self.params["step"] = value
