# -*- coding: utf-8 -*-


import time
from threading import Thread, Lock, Event
from datetime import datetime
import os
from pathlib import Path

# Paths defined for OS-independent operation
_MAIN_PATH = Path("src/")
_LOG_PATH = Path("log/")
_FILES_PATH = Path("input_files/")


class FileDataGatherer:
    """
    Gathers data from an existing text file, used for testing and as proof of concept.
    Expected input file format the same as the output log file, but with no headers line and fixed at 2 magnetometers.
    It uses the same interface with the main program as other possible DataGatherer modules.
    """
    def __init__(self, pipe):
        """
        :param pipe: Duplex pipe connecting to application core. Used for sending samples/ACKs and receiving commands.
        """
        # file = "survey_example_2Hz.txt"  # Input file
        file = "survey_example_20Hz.txt"
        n_magnetometers = 2  # Number of magnetometers

        self.filein = str(_FILES_PATH / file)
        time_now = datetime.now().strftime("%Y_%m_%d__%H_%M_%S")
        self.fileout = str(_LOG_PATH / "{}.txt".format(time_now))
        self.n_magnetometers = n_magnetometers

        # Adaptive header to number of magnetometers
        with open(self.fileout, "w") as f:
            headers = "## Columns: LONGITUDE LATITUDE "
            for m in range(self.n_magnetometers):
                headers += "READING_{} ".format(m+1)
            headers += "INT_LOCK BATTERY\n"
            f.write(headers)

        self.pipe = pipe
        self.pipe_lock = Lock()

        self.reading = Event()  # Enhanced boolean, allows blocking until set
        self.unpaused = Event()
        self.unpaused.set()

        t1 = Thread(target=self._command_thread)
        t2 = Thread(target=self._data_gather)
        t1.start()
        t2.start()
        t1.join()
        t2.join()

    def _command_thread(self):
        """
        Handles incoming commands to the Gatherer and processes them as appropriate.
        "log" (start), "pause" and "stop" are considered, any other commands are ignored.
        """

        cmd = ""

        while cmd != "stop":
            cmd = self.pipe.recv()

            if cmd == "log":
                self.start = time.time()
                self.reading.set()
                self.unpaused.set()
                self.send({"type": "ACK"})
            elif cmd == "pause":
                self.unpaused.clear()
                self.send({"type": "ACK"})
            elif cmd == "stop":
                self.reading.clear()
                self.unpaused.set()  # To prevent _data_gatherer to infinitely wait for unpause
                self.send({"type": "ACK"})

    def send(self, d):
        """
        Ensures thread-safe use of the pipe, so ACKs and data packets are not sent at the same time.

        :param d: formatted packet, meant for ACK or data.
        """
        with self.pipe_lock:
            self.pipe.send(d)

    def _data_gather(self):
        """
        Sample packet creation loop.
        To be re-coded for each Data Gatherer type
        FileDataGatherer:
            Waits for start command. Then reads lines from input file one at a time while checking for pause or stop
        commands in each iteration. Formats the dictionary sample packet and sends it for each input line. If stopped or
        EOF is reached, None is sent as a poison pill and the thread is terminated.
        """
        self.reading.wait()  # Waits until first "start" command is sent

        with open(self.filein, "r") as filein, open(self.fileout, "a") as fileout:

            self.unpaused.wait()  # Blocks until unpaused if self.paused is set
            while self.reading.is_set():
                dict_out = {"type": "sample"}

                line = filein.readline()
                line_elements = line.split(" ")

                if len(line_elements) == 1:  # Blank line may come before EOF
                    break

                # Packet components
                dict_out["longitude"] = float(line_elements[0])
                dict_out["latitude"] = float(line_elements[1])

                log_line = "{} {} ".format(dict_out["longitude"], dict_out["latitude"])

                for m in range(self.n_magnetometers):
                    reading = line_elements[2+m]
                    dict_out["reading_{}".format(m+1)] = float(reading)
                    log_line += reading+" "

                dict_out["battery"] = float(line_elements[2+self.n_magnetometers])
                dict_out["integrity_lock"] = bool(int(line_elements[3+self.n_magnetometers]))

                log_line += "{} {}\n".format(dict_out["battery"], int(dict_out["integrity_lock"]))

                fileout.write(log_line)  # Output file is updated with a new line

                self.send(dict_out)
                time.sleep(0)  # Recreates delay between samples
                self.unpaused.wait()
        self.end = time.time()
        print("Time: {}".format(self.end-self.start))
        self.send(None)  # Poison pill to signal Gatherer termination


class ReplayDataGatherer:
    """
    Test Gatherer to re-play a survey plot from the drone log file.
    Fundamentally the same as FileDataGatherer, but it is set up for a different input file format.
    Expected input file: GPS and single magnetometer, with lots of control data that need to be filtered out.

    NOTE: VERY sensitive to input file format, meant exclusively for the log file generated at the drone side.
    """
    def __init__(self, pipe):
        """
        :param pipe: Duplex pipe connecting to application core. Used for sending samples/ACKs and receiving commands.
        """
        file = "10-05-2019-10_51_42_test_flight.txt"  # Input file

        self.filein = str(_FILES_PATH / file)
        self.index_error_counter = 0  # Sanity checks, due to possible inconsistencies in input file
        self.value_error_counter = 0

        self.pipe = pipe
        self.pipe_lock = Lock()

        self.reading = Event()
        self.unpaused = Event()
        self.unpaused.set()

        Thread(target=self._command_thread).start()
        self._data_gather()

    def _command_thread(self):
        """
        Handles incoming commands to the Gatherer and processes them as appropriate.
        "log" (start), "pause" and "stop" are considered, any other commands are ignored.
        """

        cmd = ""

        while cmd != "stop":
            cmd = self.pipe.recv()

            if cmd == "log":
                self.start = time.time()
                self.reading.set()
                self.unpaused.set()
                self.send({"type": "ACK"})
            elif cmd == "pause":
                self.unpaused.clear()
                self.send({"type": "ACK"})
            elif cmd == "stop":
                self.reading.clear()
                self.unpaused.set()  # To prevent _data_gatherer to infinitely wait for unpause
                self.send({"type": "ACK"})

    def send(self, d):
        """
        Ensures thread-safe use of the pipe, so ACKs and data packets are not sent at the same time.

        :param d: formatted packet, meant for ACK or data.
        """
        with self.pipe_lock:
            self.pipe.send(d)

    def _data_gather(self):
        """
        MAIN LOOP for the Data Gatherer
        To be re-coded for each Data Gatherer type
        FileDataGatherer:
            Reads input file, formats packet dictionary
        """
        self.reading.wait()

        with open(self.filein, "r") as filein:
            last_timestamp = 0
            self.unpaused.wait()  # Blocks until unpaused

            while self.reading.is_set():  # One line represents one reading in the test file
                line = filein.readline()

                if not (115 < len(line) < 150):
                    continue  # Line is skipped if it does not have the expected length
                try:
                    gps, magnetometer = line.split("&")

                    gps_elements = gps.split(",")
                    magnet_elements = magnetometer.split(" ")
                    new_timestamp = float(gps_elements[1])

                    if new_timestamp == last_timestamp:
                        continue  # Sample is skipped if same GPS is found (GPS at 2Hz, magnetometer at 20Hz)
                    last_timestamp = new_timestamp

                    dict_out = {"type": "sample"}

                    latitude_coded = float(gps_elements[2])
                    if gps_elements[3] == "S":  # "North/South" to +/-
                        latitude_coded *= -1
                    longitude_coded = float(gps_elements[4])  # "East/West" to +/-
                    if gps_elements[5] == "W":
                        longitude_coded *= -1

                    latitude_aux = round(latitude_coded, -2)  # 100xDegrees
                    latitude = latitude_aux / 100 + (latitude_coded - latitude_aux) / 60
                    longitude_aux = round(longitude_coded, -2)
                    longitude = longitude_aux / 100 + (longitude_coded - longitude_aux) / 60

                    field = float(magnet_elements[1])
                    i_lock = bool(int(magnet_elements[2]))

                    battery = float(magnet_elements[9])

                    # Dictionary (inter-process "packet") components
                    dict_out["longitude"] = longitude
                    dict_out["latitude"] = latitude
                    dict_out["reading_1"] = field
                    dict_out["battery"] = battery
                    dict_out["integrity_lock"] = i_lock

                    self.send(dict_out)
                    self.unpaused.wait()

                except IndexError:
                    self.index_error_counter += 1
                except ValueError:
                    self.value_error_counter += 1
        # Sanity checks to look for unaccounted file inconsistencies
        print("Index Error number: " + str(self.index_error_counter))
        print("Value Error number: " + str(self.value_error_counter))
        self.send(None)  # Poison pill to signal Gatherer termination


class FifoDataGatherer:
    """
    Communicates with receiver script through FIFO files and receives data in real time.
    """
    def __init__(self, pipe):
        """
        :param pipe: Duplex pipe connecting to application core. Used for sending samples/ACKs and receiving commands.
        """
        self.pipe = pipe
        self.pipe_lock = Lock()
        time_now = datetime.now().strftime("%Y_%m_%d__%H_%M_%S")
        self.fileout = str(_LOG_PATH / "{}.txt".format(time_now))
        self.FIFO_CMD = "fifo_cmd"
        self.FIFO_DATA = "fifo_dat"

        try:
            os.mkfifo(self.FIFO_CMD)
        except OSError:
            os.remove(self.FIFO_CMD)
            os.mkfifo(self.FIFO_CMD)

        try:
            os.mkfifo(self.FIFO_DATA)
        except OSError:
            os.remove(self.FIFO_DATA)
            os.mkfifo(self.FIFO_DATA)

        with open(self.fileout, "w") as f:
            headers = "## Columns: LONGITUDE LATITUDE [MAGNETOMETER_1, MAGNETOMETER_2, ...] INT_LOCK BATTERY\n"
            f.write(headers)

        # Thread(target=self._data_thread, daemon=False).start()
        t1 = Thread(target=self._init_receiver, daemon=True).start()  # Separate thread for the receiver C++ script
        t2 = Thread(target=self._command_thread, daemon=True).start()
        # self._command_thread()
        self._data_thread()
        print("Starting joining")
        t1.join()
        print("Init receiver joined")
        t2.join()
        print("Command thread joined")

    def _command_thread(self):
        """
        Handles incoming commands to the Gatherer and processes them as appropriate.
        "log" (start), "pause" and "stop" are considered, any other commands are ignored.
        """

        print("[Data Gatherer] Starting sending thread")
        cmd = ""
        while cmd != "stop":
            cmd = self.pipe.recv()
            print("[Data Gatherer] Command will be sent - " + cmd)
            with open(self.FIFO_CMD, "w") as cmd_file:
                cmd_file.write(cmd)
            print("[Data Gatherer] Command has been sent - "+cmd)

        print("[Data Gatherer] Command thread was terminated")
        os.remove(self.FIFO_CMD)
        print("command thread ded")

    def _data_thread(self):
        print("[Data Gatherer] Data thread started")
        with open(self.fileout, "w") as fileout:
            reading = True
            while reading:
                with open(self.FIFO_DATA, "r") as data_file:
                    line = data_file.readline()
                print("[Data Gatherer] Line has been read")
                try:
                    line = line.strip()
                    if "None" in line:
                        reading = False
                        print("[Data Gatherer] None received - terminating")
                        break
                    elif "acknowledge" in line:
                        dict_out = {"type": "ACK"}
                        print("[Data Gatherer] ACK received")
                    else:
                        dict_out = {"type": "sample"}

                        print("[DataGatherer] Data sample received" + line)

                        line_elements = line.split(",")
                        gps_data = line_elements[0:3]  # GPS[0] is timestamp, unnecessary for now
                        latitude_coded = float(gps_data[1])  # float version of DDMM.MMMM
                        longitude_coded = float(gps_data[2])  # float version of DDDMM.MMMM

                        latitude_aux = round(latitude_coded, -2)  # 100xDegrees
                        latitude = latitude_aux/100 + (latitude_coded-latitude_aux)/60
                        longitude_aux = round(longitude_coded, -2)
                        longitude = longitude_aux/100 + (longitude_coded-longitude_aux)/60

                        dict_out["latitude"] = latitude
                        dict_out["longitude"] = longitude

                        dict_out["battery"] = abs(float(line_elements[4]))

                        n_magnetometers = (len(line_elements)-3) // 2  # 3 fields for gps. +2 fields per magnetometer

                        i_lock = True
                        logged_sample = "{} {}".format(dict_out["longitude"], dict_out["latitude"])  # log entry

                        for m in range(n_magnetometers):
                            dict_out["reading_{}".format(m+1)] = float(line_elements[3+2*m])
                            i_lock = (i_lock and (float(line_elements[4 + 2*m]) > 0))  # i_lock covers all magnetometers
                            logged_sample += " {}".format(dict_out["reading_{}".format(m+1)])
                        dict_out["integrity_lock"] = i_lock

                        logged_sample += " {} {}\n".format(dict_out["battery"], int(dict_out["integrity_lock"]))
                        fileout.write(logged_sample)

                    print(dict_out)
                    self.send(dict_out)
                except IndexError:
                        print("BAD SAMPLE!")

        self.send(None)

    def delete_fifos(self):
        """
        Unimplemented.
        Deletes FIFO files. To be called always at program shutdown.
        """
        try:
            os.remove(self.FIFO_DATA)
            os.remove(self.FIFO_CMD)
        except OSError:
            pass

    def send(self, d):
        """
        Ensures thread-safe use of the pipe, so ACKs and data packets are not sent at the same time.

        :param d: formatted packet, meant for ACK or data.
        """
        with self.pipe_lock:
            self.pipe.send(d)

    def _init_receiver(self):
        """
        To be called on a separate thread/process.
        Launches the MagPro utility to handle the radio communications and gives it control of the thread.
        MagPro must be launched manually at the aircraft side.
        """
        os.system(str(_MAIN_PATH / "MagPro" / "MagPro"))
